/**
 *
 * $Id: Visual.c,v 1.19 1998/09/16 19:31:47 danny Exp $
 *
 * Copyright (C) 1995 Free Software Foundation, Inc.
 *
 * This file is part of the GNU LessTif Library.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 **/

static const char rcsid[] = "$Id: Visual.c,v 1.19 1998/09/16 19:31:47 danny Exp $";

#include <LTconfig.h>
#include <XmI/XmI.h>
#include <stdio.h>
#include <limits.h>

#include <Xm/XmP.h>
#include <Xm/PrimitiveP.h>
#include <Xm/ManagerP.h>
#include <Xm/ExtObjectP.h>
#include <Xm/GadgetP.h>

#include <XmI/DebugUtil.h>

#define	COLOR_CACHE_SIZE		5

#define	PCT_BRIGHTNESS			(6 * 0xffff / 100)

/* How much lighter/darker to make things in default routine */

#define	PCT_DARK_BOTTOM		70	/* lighter (less dark, actually) */
#define	PCT_DARK_TOP		50	/* lighter */
#define	PCT_LIGHT_BOTTOM	55	/* darker */
#define	PCT_LIGHT_TOP		80	/* darker */
#define	PCT_MEDIUM_BOTTOM_BASE	40	/* darker */
#define	PCT_MEDIUM_BOTTOM_RANGE	25
#define	PCT_MEDIUM_TOP_BASE	60	/* lighter */
#define	PCT_MEDIUM_TOP_RANGE	-30
#define	PCT_SELECT		85	/* darker (lighter if "dark") */

static XmColorProc _color_proc = NULL;
static XmColorData color_cache[COLOR_CACHE_SIZE];

/*
 *  Note: 
 *  
 * At one time there was no cache in the color routines. When displaying
 * remotely, profile data indicated that the vast majority of the time
 * spent in _XtCreateWidget was spent getting colors from these routines.
 * In Aug 1997, Jon Christopher (jac8792@tamu.edu) added a cache to each
 * of these routines so that it will remember the data from the last call
 * and avoid server round trips if they are called again under the same
 * conditions.  These changes speeded up _XtCreateWidget by 10 times.
 *
 *  More recently (Sept 1998), Jamie Gritton (jamie@gritton.org) improved
 *  the cache by using the proper _Xm* routines as defined in XmP.h.
 *  Jamie has also implemented routines which use better default color
 *  calculations based on Motif-like empirical brightness calculations.
 *  Thanks, Jamie!  */


static void color_fail_warn(void)
{
  static int warned = False;
  
  if (!warned) {
    warned=True;
    _XmWarning(NULL,
	       "Cannot allocate colormap entry, some colors may be incorrect");
  }
}

static void _XmColorProcDefaultProc(XColor *bg_color, XColor *fg_color,
				    XColor *sel_color, XColor *ts_color,
				    XColor *bs_color);

XmColorProc
XmSetColorCalculation(XmColorProc proc)
{
    XmColorProc ret_val;

    ret_val = _color_proc;

    if (proc == NULL)
    {
	_color_proc = _XmColorProcDefaultProc;
    }
    else
    {
	_color_proc = proc;
    }

    return ret_val;
}

XmColorProc
XmGetColorCalculation()
{
    return _color_proc;
}

/* Get a color cache entry */

XmColorData *
_XmGetColors(Screen *screen, Colormap color_map, Pixel background)
{
    XmColorData cd;
    XmColorData *cc;

    /* Construct a new blank record to possibly add */

    cd.allocated = 0;
    cd.screen = screen;
    cd.color_map = color_map;
    cd.background.pixel = background;
    cd.foreground.pixel =
	cd.top_shadow.pixel =
	cd.bottom_shadow.pixel =
	cd.select.pixel = 0;

    /* Get it from the cache, putting it in if necessary */

    cc = _XmAddToColorCache(&cd);

    /* Call the color proc if the entry is new */

    if (!(cc->allocated & XmBACKGROUND))
    {
	cc->allocated |= XmBACKGROUND;

	XQueryColor(XDisplayOfScreen(screen),
		    color_map,
		    &cc->background);

	_color_proc(&cc->background,
		    &cc->foreground,
		    &cc->select,
		    &cc->top_shadow,
		    &cc->bottom_shadow);
    }

    return cc;
}

/* Search the color cache for a matching entry */

Boolean
_XmSearchColorCache(unsigned int which, XmColorData *values,
		    XmColorData **ret)
{
    XmColorData *cc;

    for (cc = color_cache; cc < color_cache + COLOR_CACHE_SIZE; cc++)
	if (cc->allocated &&
	    (!(which & XmLOOK_AT_SCREEN) ||
	     cc->screen == values->screen) &&
	    (!(which & XmLOOK_AT_CMAP) ||
	     cc->color_map == values->color_map) &&
	    (!(which & XmLOOK_AT_BACKGROUND) ||
	     cc->background.pixel == values->background.pixel) &&
	    (!(which & XmLOOK_AT_FOREGROUND) ||
	     ((cc->allocated & XmFOREGROUND) &&
	      cc->foreground.pixel == values->foreground.pixel)) &&
	    (!(which & XmLOOK_AT_TOP_SHADOW) ||
	     ((cc->allocated & XmTOP_SHADOW) &&
	      cc->top_shadow.pixel == values->top_shadow.pixel)) &&
	    (!(which & XmLOOK_AT_BOTTOM_SHADOW) ||
	     ((cc->allocated & XmBOTTOM_SHADOW) &&
	      cc->bottom_shadow.pixel == values->bottom_shadow.pixel)) &&
	    (!(which & XmLOOK_AT_SELECT) ||
	     ((cc->allocated & XmSELECT) &&
	      cc->select.pixel == values->select.pixel)))
	{
	    /* Found the entry.  Bump it to the front of the cache
	     * to find it quicker next time, and to make this an LRU cache.
	     * Kind of expensive perhaps, but for a small cache that's fine.
	     */

	    if (cc > color_cache)
	    {
		XmColorData temp_cd = *cc;
		memmove(color_cache + 1, color_cache,
			(char *)cc - (char *)color_cache);
		*color_cache = temp_cd;
	    }
	    *ret = color_cache;
	    return True;
	}

    *ret = NULL;
    return False;
}

/* Add an entry to the color cache, or augment an existing entry */

XmColorData *
_XmAddToColorCache(XmColorData *new_rec)
{
    XmColorData *cc;

    if (_XmSearchColorCache(XmLOOK_AT_SCREEN | XmLOOK_AT_CMAP |
			    XmLOOK_AT_BACKGROUND, new_rec, &cc))
    {
	/* Entry exists - add new data */

	cc->allocated |= new_rec->allocated;
	if (new_rec->allocated & XmBACKGROUND)
	{
	    cc->background = new_rec->background;
	}
	if (new_rec->allocated & XmFOREGROUND)
	{
	    cc->foreground = new_rec->foreground;
	}
	if (new_rec->allocated & XmTOP_SHADOW)
	{
	    cc->top_shadow = new_rec->top_shadow;
	}
	if (new_rec->allocated & XmBOTTOM_SHADOW)
	{
	    cc->bottom_shadow = new_rec->bottom_shadow;
	}
	if (new_rec->allocated & XmSELECT)
	{
	    cc->select = new_rec->select;
	}
	return cc;
    }
    else
    {
	/* Key doesn't exist - add the entire entry */

	memmove(color_cache + 1, color_cache,
		(COLOR_CACHE_SIZE - 1) * sizeof( XmColorData));
	color_cache[0] = *new_rec;
	return color_cache;
    }
}

/* Get a field from a color cache entry, allocating it if necessary */

Pixel
_XmAccessColorData(XmColorData *cd, unsigned char which)
{
    /* If allocation fails, punt to back or white.  Perhaps I should use an
     * excerpt from _XmColorProcDefaultProc, but that hardly seems worthwhile
     * for an emergency.  Likewise the potential redundancy.
     */

    switch(which)
    {
    case XmBACKGROUND:
	return cd->background.pixel;
    case XmFOREGROUND:
	if (!(cd->allocated & XmFOREGROUND))
	{
	    cd->allocated |= XmFOREGROUND;
	    if (!XAllocColor(XDisplayOfScreen(cd->screen),
			     cd->color_map,
			     &cd->foreground))
	    {
		DEBUGOUT(XdbDebug0(__FILE__, NULL, "%s:_XmAccessColorData(%d) - Could not allocate XmFOREGROUND\n",
				   __FILE__,__LINE__));
		color_fail_warn();
		cd->foreground.pixel = (unsigned long) cd->background.red +
		    cd->background.blue + cd->background.green >= 0x18000 ?
		    _XmBlackPixel(cd->screen, cd->color_map, cd->foreground) :
		    _XmWhitePixel(cd->screen, cd->color_map, cd->foreground);
	    }
	}
	return cd->foreground.pixel;
    case XmTOP_SHADOW:
	if (!(cd->allocated & XmTOP_SHADOW))
	{
	    cd->allocated |= XmTOP_SHADOW;
	    if (!XAllocColor(XDisplayOfScreen(cd->screen),
			     cd->color_map,
			     &cd->top_shadow))
	    {
		DEBUGOUT(XdbDebug0(__FILE__, NULL, "%s:_XmAccessColorData(%d) - Could not allocate XmTOP_SHADOW\n",
				   __FILE__,__LINE__));
		color_fail_warn();
		cd->top_shadow.pixel = (unsigned long) cd->background.red +
		    cd->background.blue + cd->background.green >= 0x18000 ?
		    _XmBlackPixel(cd->screen, cd->color_map, cd->top_shadow) :
		    _XmWhitePixel(cd->screen, cd->color_map, cd->top_shadow);
	    }
	}
	return cd->top_shadow.pixel;
    case XmBOTTOM_SHADOW:
	if (!(cd->allocated & XmBOTTOM_SHADOW))
	{
	    cd->allocated |= XmBOTTOM_SHADOW;
	    if (!XAllocColor(XDisplayOfScreen(cd->screen),
			     cd->color_map,
			     &cd->bottom_shadow))
	    {
		DEBUGOUT(XdbDebug0(__FILE__, NULL, "%s:_XmAccessColorData(%d) - Could not allocate XmBOTTOM_SHADOW\n",
				   __FILE__,__LINE__));
		color_fail_warn();
		cd->bottom_shadow.pixel = (unsigned long) cd->background.red +
		    cd->background.blue + cd->background.green >= 0x18000 ?
		    _XmBlackPixel(cd->screen, cd->color_map,
				  cd->bottom_shadow) :
		    _XmWhitePixel(cd->screen, cd->color_map,
				  cd->bottom_shadow);
	    }
	}
	return cd->bottom_shadow.pixel;
    case XmSELECT:
	if (!(cd->allocated & XmSELECT))
	{
	    cd->allocated |= XmSELECT;
	    if (!XAllocColor(XDisplayOfScreen(cd->screen),
			     cd->color_map,
			     &cd->select))
	    {
		DEBUGOUT(XdbDebug0(__FILE__, NULL, "%s:_XmAccessColorData(%d) - Could not allocate XmSELECT\n",
				   __FILE__,__LINE__));
		color_fail_warn();
		cd->select.pixel = (unsigned long) cd->background.red +
		    cd->background.blue + cd->background.green >= 0x18000 ?
		    _XmBlackPixel(cd->screen, cd->color_map, cd->select) :
		    _XmWhitePixel(cd->screen, cd->color_map, cd->select);
	    }
	}
	return cd->select.pixel;
    }

    _XmWarning(NULL, "Invalid color specification in _XmAccessColorData");
    return 0;
}

/* External entry point - get requested colors for a widget */

void
XmGetColors(Screen *screen,
	    Colormap color_map,
	    Pixel background,
	    Pixel *foreground_ret,
	    Pixel *top_shadow_ret,
	    Pixel *bottom_shadow_ret,
	    Pixel *select_ret)
{
    XmColorData *cd;

    cd = _XmGetColors( screen, color_map, background);

    if (foreground_ret)
	*foreground_ret = _XmAccessColorData( cd, XmFOREGROUND);
    if (top_shadow_ret)
	*top_shadow_ret = _XmAccessColorData( cd, XmTOP_SHADOW);
    if (bottom_shadow_ret)
	*bottom_shadow_ret = _XmAccessColorData( cd, XmBOTTOM_SHADOW);
    if (select_ret)
	*select_ret = _XmAccessColorData( cd, XmSELECT);
}

void
XmChangeColor(Widget widget,
	      Pixel background)
{
    Pixel foreground;
    Pixel top_shadow;
    Pixel bottom_shadow;
    Pixel select;

    XmGetColors(XtScreenOfObject(widget),
		ColormapOfObject(widget),
		background,
		&foreground,
		&top_shadow,
		&bottom_shadow,
		&select);

    XtVaSetValues(widget,
		  XmNbackground, background,
		  XmNforeground, foreground,
		  XmNtopShadowColor, top_shadow,
		  XmNbottomShadowColor, bottom_shadow,
		  XmNarmColor, select,
		  XmNselectColor, select,
		  XmNtroughColor, select,
		  NULL);
}

/* the default color calculation */
static void
_XmColorProcDefaultProc(XColor *bg_color,
			XColor *fg_color,
			XColor *sel_color,
			XColor *ts_color,
			XColor *bs_color)
{
    long brightness, brightadj;
    XColor _widgetBackground;

    _widgetBackground = *bg_color;

    /* Find the "brightness" of an RGB color.  The "right" way seems to use
     * empirical values like the default thresholds below, but it boils down
     * to red is twice as bright as blue and green is thrice blue.
     */

    brightness = _widgetBackground.red + _widgetBackground.red +
	_widgetBackground.green + _widgetBackground.green +
	_widgetBackground.green + _widgetBackground.blue;

    if (fg_color)
	fg_color->red = fg_color->green = fg_color->blue =
	    brightness > XmDEFAULT_FOREGROUND_THRESHOLD * PCT_BRIGHTNESS
	    ? 0 : 0xffff;

    /* For "dark" backgrounds, make everything a fixed %age lighter */

    if (brightness < XmDEFAULT_DARK_THRESHOLD * PCT_BRIGHTNESS)
    {
	if (sel_color)
	{
	    sel_color->red = 0xffff -
		((0xffff - _widgetBackground.red) * PCT_SELECT + 50) / 100;
	    sel_color->green = 0xffff -
		((0xffff - _widgetBackground.green) * PCT_SELECT + 50) / 100;
	    sel_color->blue = 0xffff -
		((0xffff - _widgetBackground.blue) * PCT_SELECT + 50) / 100;
	}
	if (ts_color)
	{
	    ts_color->red = 0xffff -
		((0xffff - _widgetBackground.red) * PCT_DARK_TOP + 50) / 100;
	    ts_color->green = 0xffff -
		((0xffff - _widgetBackground.green) * PCT_DARK_TOP + 50) / 100;
	    ts_color->blue = 0xffff -
		((0xffff - _widgetBackground.blue) * PCT_DARK_TOP + 50) / 100;
	}
	if (bs_color)
	{
	    bs_color->red = 0xffff -
		((0xffff - _widgetBackground.red) * PCT_DARK_BOTTOM) / 100;
	    bs_color->green = 0xffff - ((0xffff - _widgetBackground.green) *
					PCT_DARK_BOTTOM + 50) / 100;
	    bs_color->blue = 0xffff - ((0xffff - _widgetBackground.blue) *
				       PCT_DARK_BOTTOM + 50) / 100;
	}
    }

    /* For "light" background, make everything a fixed %age darker */

    else if (brightness > XmDEFAULT_LIGHT_THRESHOLD * PCT_BRIGHTNESS)
    {
	if (sel_color)
	{
	    sel_color->red = (_widgetBackground.red * PCT_SELECT + 50) / 100;
	    sel_color->green =
		(_widgetBackground.green * PCT_SELECT + 50) / 100;
	    sel_color->blue = (_widgetBackground.blue * PCT_SELECT + 50) / 100;
	}
	if (ts_color)
	{
	    ts_color->red = (_widgetBackground.red * PCT_LIGHT_TOP + 50) / 100;
	    ts_color->green =
		(_widgetBackground.green * PCT_LIGHT_TOP + 50) / 100;
	    ts_color->blue =
		(_widgetBackground.blue * PCT_LIGHT_TOP + 50) / 100;
	}
	if (bs_color)
	{
	    bs_color->red =
		(_widgetBackground.red * PCT_LIGHT_BOTTOM + 50) / 100;
	    bs_color->green =
		(_widgetBackground.green * PCT_LIGHT_BOTTOM + 50) / 100;
	    bs_color->blue =
		(_widgetBackground.blue * PCT_LIGHT_BOTTOM + 50) / 100;
	}
    }

    /* For "medium" background, select is a fixed %age darker;
     * top (lighter) and bottom (darker) are a variable %age
     * based on the background's brightness
     */

    else
    {
	brightness = (brightness + (PCT_BRIGHTNESS >> 1)) / PCT_BRIGHTNESS;
	if (sel_color)
	{
	    sel_color->red = (_widgetBackground.red * PCT_SELECT + 50) / 100;
	    sel_color->green =
		(_widgetBackground.green * PCT_SELECT + 50) / 100;
	    sel_color->blue =
		(_widgetBackground.blue * PCT_SELECT + 50) / 100;
	}
	if (ts_color)
	{
	    brightadj = PCT_MEDIUM_TOP_BASE +
		(brightness * PCT_MEDIUM_TOP_RANGE + 50) / 100;
	    ts_color->red = 0xffff -
		((0xffff - _widgetBackground.red) * brightadj + 50) / 100;
	    ts_color->green = 0xffff -
		((0xffff - _widgetBackground.green) * brightadj + 50) / 100;
	    ts_color->blue = 0xffff -
		((0xffff - _widgetBackground.blue) * brightadj + 50) / 100;
	}
	if (bs_color)
	{
	    brightadj = PCT_MEDIUM_BOTTOM_BASE +
		(brightness * PCT_MEDIUM_BOTTOM_RANGE + 50) / 100;
	    bs_color->red = (_widgetBackground.red * brightadj + 50) / 100;
	    bs_color->green = (_widgetBackground.green * brightadj + 50) / 100;
	    bs_color->blue = (_widgetBackground.blue * brightadj + 50) / 100;
	}
    }
}

/****************************************/
/* private functions for default colors */
/****************************************/

/* I suspect these functions (__XmRGB2HSV and __XMHSV2RGB) shouldn't be
 * exported, and in fact aren't ever used at all (they're not longer used
 * here).  They'll stick around for now just in case someone needs them.  JHG
 *
 * Alexander Mai pointed out that mwm needs them, so I put them back in. -jac
 */


#define FLOOR(x) ((int)((x) - 0.5) > (x) ? (x) : (x)-1)

/* Convert between RGB and HSV */
void
__XmRGB2HSV(unsigned short r,
	    unsigned short g,
	    unsigned short b,
	    double *hr,
	    double *sr,
	    double *vr)
{
    double rd, gd, bd, max, min, del;
    double rc, gc, bc;
    double v, h, s;

    /* convert RGB to HSV */
    rd = r / 65535.0;		/* rd,gd,bd range 0-1 instead of 0-65535 */
    gd = g / 65535.0;
    bd = b / 65535.0;

    /* compute maximum of rd,gd,bd */
    if (rd >= gd)
    {
	if (rd >= bd)
	{
	    max = rd;
	}
	else
	{
	    max = bd;
	}
    }
    else
    {
	if (gd >= bd)
	{
	    max = gd;
	}
	else
	{
	    max = bd;
	}
    }

    /* compute minimum of rd,gd,bd */
    if (rd <= gd)
    {
	if (rd <= bd)
	{
	    min = rd;
	}
	else
	{
	    min = bd;
	}
    }
    else
    {
	if (gd <= bd)
	{
	    min = gd;
	}
	else
	{
	    min = bd;
	}
    }


    del = max - min;
    v = max;
    if (max != 0.0)
    {
	s = (del) / max;
    }
    else
    {
	s = 0.0;
    }

    h = -1;
    if (s != 0.0)
    {
	rc = (max - rd) / del;
	gc = (max - gd) / del;
	bc = (max - bd) / del;

	if (rd == max)
	{
	    h = bc - gc;
	}
	else if (gd == max)
	{
	    h = 2 + rc - bc;
	}
	else if (bd == max)
	{
	    h = 4 + gc - rc;
	}

	h = h * 60;
	if (h < 0)
	    h += 360;
    }

    *hr = h;
    *sr = s;
    *vr = v;
}

/* Convert between HSV and RGB */
void
_XmHSV2RGB(double h,
	   double s,
	   double v,
	   unsigned short *rr,
	   unsigned short *gr,
	   unsigned short *br)
{
    int j;
    double rd, gd, bd;
    double f, p, q, t;

    /* convert HSV back to RGB */
    if (h == -1 || s == 0.0)
    {
	rd = v;
	gd = v;
	bd = v;
    }
    else
    {
	if (h == 360.0)
	{
	    h = 0.0;
	}
	h = h / 60.0;
	j = (int)FLOOR(h);
	if (j < 0)
	{
	    j = 0;		/* either h or floor seem to go neg on some sys */
	}
	f = h - j;
	p = v * (1 - s);
	q = v * (1 - (s * f));
	t = v * (1 - (s * (1 - f)));

	switch (j)
	{
	case 0:
	    rd = v;
	    gd = t;
	    bd = p;
	    break;
	case 1:
	    rd = q;
	    gd = v;
	    bd = p;
	    break;
	case 2:
	    rd = p;
	    gd = v;
	    bd = t;
	    break;
	case 3:
	    rd = p;
	    gd = q;
	    bd = v;
	    break;
	case 4:
	    rd = t;
	    gd = p;
	    bd = v;
	    break;
	case 5:
	    rd = v;
	    gd = p;
	    bd = q;
	    break;
	default:
	    rd = v;
	    gd = t;
	    bd = p;
	    break;		/* never happen */
	}
    }


    *rr = (unsigned short)FLOOR((rd * 65535.0) + 0.5);
    *gr = (unsigned short)FLOOR((gd * 65535.0) + 0.5);
    *br = (unsigned short)FLOOR((bd * 65535.0) + 0.5);
}


void
_XmForegroundColorDefault(Widget w,
			  int offset,
			  XrmValue *val)
{
    static Pixel foreground;

    val->addr = (XPointer)&foreground;
    val->size = sizeof foreground;
    XmGetColors(XtScreenOfObject(w), ColormapOfObject(w), XmIsGadget(w) ?
		XmParentBackground(w) : XtBackground(w), &foreground, NULL,
		NULL, NULL);
}

void
_XmHighlightColorDefault(Widget w,
			 int offset,
			 XrmValue *val)
{
    static Pixel highlight;

    val->addr = (XPointer)&highlight;
    val->size = sizeof highlight;
    XmGetColors(XtScreenOfObject(w), ColormapOfObject(w), XmIsGadget(w) ?
		XmParentBackground(w) : XtBackground(w), &highlight, NULL,
		NULL, NULL);
}

void
_XmBackgroundColorDefault(Widget w,
			  int offset,
			  XrmValue *val)
{
    Colormap colormap = CoreColormap(w);
    static XColor backgroundColor;
    static int inited = False;
    static Colormap last_colormap;
    static Display *last_display;

    /* check cache */
    if (inited && 
	last_colormap == colormap &&
	last_display == XtDisplayOfObject(w))
    {
	val->addr = (XtPointer)&backgroundColor.pixel;
	return;
    }
    else
    {
	inited = True;
	last_colormap = colormap;
	last_display = XtDisplayOfObject(w);
    }

    if (!XParseColor(XtDisplayOfObject(w),
		     colormap,
		     "rgb:72/9f/ff",
		     &backgroundColor) ||
	!XAllocColor(XtDisplayOfObject(w),
		     colormap,
		     &backgroundColor))
    {
	DEBUGOUT(XdbDebug0(__FILE__, NULL, "%s:_XmBackgroundColorDefault(%d) - Could not allocate color\n",
		__FILE__,__LINE__));
	color_fail_warn();
	backgroundColor.pixel = _XmWhitePixelOfObject(w);
    }

    val->addr = (XtPointer)&backgroundColor.pixel;
}


void
_XmTopShadowColorDefault(Widget w,
			 int offset,
			 XrmValue *val)
{
    static Pixel top_shadow;

    val->addr = (XPointer)&top_shadow;
    val->size = sizeof top_shadow;
    XmGetColors(XtScreenOfObject(w), ColormapOfObject(w), XmIsGadget(w) ?
		XmParentBackground(w) : XtBackground(w), NULL, &top_shadow,
		NULL, NULL);
}


void
_XmBottomShadowColorDefault(Widget w,
			    int offset,
			    XrmValue *val)
{
    static Pixel bottom_shadow;

    val->addr = (XPointer)&bottom_shadow;
    val->size = sizeof bottom_shadow;
    XmGetColors(XtScreenOfObject(w), ColormapOfObject(w), XmIsGadget(w) ?
		XmParentBackground(w) : XtBackground(w), NULL, NULL,
		&bottom_shadow, NULL);
}

void
_XmSelectColorDefault(Widget w,
		      int offset,
		      XrmValue *val)
{
    static Pixel select;

    val->addr = (XPointer)&select;
    val->size = sizeof select;
    XmGetColors(XtScreenOfObject(w), ColormapOfObject(w), XmIsGadget(w) ?
		XmParentBackground(w) : XtBackground(w), NULL, NULL, NULL,
		&select);
}

/*
 * conversion args for the pixmap converters.  Thanks to Gilles Seguin
 * for pointing out the lack.
 */
#define POFFSET(field) \
    (XtPointer)XtOffsetOf(XmPrimitiveRec, field)
#define MOFFSET(field) \
    (XtPointer)XtOffsetOf(XmManagerRec, field)

/* background for everyone */
static XtConvertArgRec backgroundArgs[] =
{
    {
	XtBaseOffset, (XtPointer)XtOffsetOf(CoreRec, core.self),
	sizeof(WidgetRec *)
    }
};


/* primitive foreground */
static XtConvertArgRec PrimForegroundPixmapArgs[] =
{
    {
	XtBaseOffset, POFFSET(core.screen), sizeof(Screen *)
    },
    {
	XtBaseOffset, POFFSET(primitive.foreground), sizeof(Pixel)
    },
    {
	XtBaseOffset, POFFSET(core.background_pixel), sizeof(Pixel)
    },
    {
	XtBaseOffset, POFFSET(core.depth), sizeof(Cardinal)
    }
};

/* primitive highlight */
static XtConvertArgRec PrimHighlightPixmapArgs[] =
{
    {
	XtBaseOffset, POFFSET(core.screen), sizeof(Screen *)
    },
    {
	XtBaseOffset, POFFSET(primitive.highlight_color), sizeof(Pixel)
    },
    {
	XtBaseOffset, POFFSET(core.background_pixel), sizeof(Pixel)
    },
    {
	XtBaseOffset, POFFSET(core.depth), sizeof(Cardinal)
    }
};

/* primitive topShadow */
static XtConvertArgRec PrimTopShadowPixmapArgs[] =
{
    {
	XtBaseOffset, POFFSET(core.screen), sizeof(Screen *)
    },
    {
	XtBaseOffset, POFFSET(primitive.highlight_color), sizeof(Pixel)
    },
    {
	XtBaseOffset, POFFSET(core.background_pixel), sizeof(Pixel)
    },
    {
	XtBaseOffset, POFFSET(core.depth), sizeof(Cardinal)
    }
};

/* primitive bottomShadow */
static XtConvertArgRec PrimBottomShadowPixmapArgs[] =
{
    {
	XtBaseOffset, POFFSET(core.screen), sizeof(Screen *)
    },
    {
	XtBaseOffset, POFFSET(primitive.bottom_shadow_color), sizeof(Pixel)
    },
    {
	XtBaseOffset, POFFSET(core.background_pixel), sizeof(Pixel)
    },
    {
	XtBaseOffset, POFFSET(core.depth), sizeof(Cardinal)
    }
};

/* manager foreground */
static XtConvertArgRec ManForegroundPixmapArgs[] =
{
    {
	XtWidgetBaseOffset, MOFFSET(core.screen), sizeof(Screen *)
    },
    {
	XtWidgetBaseOffset, MOFFSET(manager.foreground), sizeof(Pixel)
    },
    {
	XtWidgetBaseOffset, MOFFSET(core.background_pixel), sizeof(Pixel)
    },
    {
	XtWidgetBaseOffset, MOFFSET(core.depth), sizeof(Cardinal)
    }
};

/* manager highlight */
static XtConvertArgRec ManHighlightPixmapArgs[] =
{
    {
	XtWidgetBaseOffset, MOFFSET(core.screen), sizeof(Screen *)
    },
    {
	XtWidgetBaseOffset, MOFFSET(manager.highlight_color), sizeof(Pixel)
    },
    {
	XtWidgetBaseOffset, MOFFSET(core.background_pixel), sizeof(Pixel)
    },
    {
	XtBaseOffset, MOFFSET(core.depth), sizeof(Cardinal)
    }
};

/* manager topShadow */
static XtConvertArgRec ManTopShadowPixmapArgs[] =
{
    {
	XtWidgetBaseOffset, MOFFSET(core.screen), sizeof(Screen *)
    },
    {
	XtWidgetBaseOffset, MOFFSET(manager.top_shadow_color), sizeof(Pixel)
    },
    {
	XtWidgetBaseOffset, MOFFSET(core.background_pixel), sizeof(Pixel)
    },
    {
	XtBaseOffset, MOFFSET(core.depth), sizeof(Cardinal)
    }
};

/* manager bottomShadow */
static XtConvertArgRec ManBottomShadowPixmapArgs[] =
{
    {
	XtWidgetBaseOffset, MOFFSET(core.screen), sizeof(Screen *)
    },
    {
	XtWidgetBaseOffset, MOFFSET(manager.bottom_shadow_color), sizeof(Pixel)
    },
    {
	XtWidgetBaseOffset, MOFFSET(core.background_pixel), sizeof(Pixel)
    },
    {
	XtBaseOffset, MOFFSET(core.depth), sizeof(Cardinal)
    }
};

/* gadget pixmap */
static XtConvertArgRec gadgetPixmapArgs[] =
{
    {
	XtBaseOffset, (XtPointer)XtOffsetOf(XmGadgetRec, object.parent),
	sizeof(WidgetRec *)
    }
};

static char *background_pixmap_name = NULL;

char *
_XmGetBGPixmapName()
{
    return background_pixmap_name;
}

void
_XmClearBGPixmap( void )
{
    background_pixmap_name = NULL;
}

static Boolean
CvtStringToBackgroundPixmap(Display *display,
			    XrmValuePtr args, Cardinal *num_args,
			    XrmValuePtr fromVal, XrmValuePtr toVal,
			    XtPointer *converter_data)
{
    static Pixmap buf;

    DEBUGOUT(XdbDebug(__FILE__, NULL,
		      "CvtStringToBackgroundPixmap: %s\n",
		      fromVal->addr));

    if (!XtIsShell((Widget)args[0].addr))
    {
	background_pixmap_name = (String)fromVal->addr;
    }

    if (toVal->addr != NULL)
    {
	if( toVal->size < sizeof(Pixmap))
	{
	    toVal->size = sizeof(Pixmap);

	    return False;
	}
	else
	{
	    *((Pixmap *)(toVal->addr)) = XmUNSPECIFIED_PIXMAP;
	    toVal->size = sizeof(Pixmap);

	    return True;
	}
    }
    else
    {
	buf = XmUNSPECIFIED_PIXMAP;
	toVal->addr = (XtPointer)&buf;
	toVal->size = sizeof(Pixmap);

	return True;
    }
}

static void
BuildPixmap(int ncargs, char *which,
	    XrmValuePtr args, Cardinal *num_args,
	    XrmValuePtr fromVal, XrmValuePtr toVal)
{
    static Pixmap pixmap;
    Screen *screen;
    Pixel foreground;
    Pixel background;
    Cardinal depth;
    char *name = fromVal->addr;

    pixmap = XmUNSPECIFIED_PIXMAP;

    toVal->addr = (char *)&pixmap;
    toVal->size = sizeof(Pixmap);

    if (*num_args != ncargs)
    {
	XtWarningMsg("wrongParameters", which, "XmToolkitError",
		     "String to Pixmap conversion needs four argument",
		     (String *)NULL, (Cardinal *)NULL);

	return;
    }

    if (strcmp(name, "unspecified_pixmap") == 0)
    {
	return;
    }

    screen = *((Screen **)args[0].addr);
    foreground = *((Pixel *)args[1].addr);
    background = *((Pixel *)args[2].addr);
    depth = *((Cardinal *)args[3].addr);

    pixmap = XmGetPixmapByDepth(screen, name,
				foreground, background, depth);
}

void
_XmCvtStringToPrimForegroundPixmap(XrmValuePtr args, Cardinal *num_args,
				   XrmValuePtr fromVal, XrmValuePtr toVal)
{
    DEBUGOUT(XdbDebug(__FILE__, NULL, "_XmCvtStringToPrimForegroundPixmap\n"));

    BuildPixmap(XtNumber(PrimForegroundPixmapArgs),
		"_XmCvtStringToPrimForegroundPixmap",
		args, num_args, fromVal, toVal);
}

void
_XmCvtStringToPrimHighlightPixmap(XrmValuePtr args, Cardinal *num_args,
				  XrmValuePtr fromVal, XrmValuePtr toVal)
{
    DEBUGOUT(XdbDebug(__FILE__, NULL, "_XmCvtStringToPrimHighlightPixmap\n"));

    BuildPixmap(XtNumber(PrimHighlightPixmapArgs),
		"_XmCvtStringToPrimHighlightPixmap",
		args, num_args, fromVal, toVal);
}

void
_XmCvtStringToPrimTopShadowPixmap(XrmValuePtr args, Cardinal *num_args,
				  XrmValuePtr fromVal, XrmValuePtr toVal)
{
    DEBUGOUT(XdbDebug(__FILE__, NULL, "_XmCvtStringToPrimTopShadowPixmap\n"));

    BuildPixmap(XtNumber(PrimTopShadowPixmapArgs),
		"_XmCvtStringToPrimTopShadowPixmap",
		args, num_args, fromVal, toVal);
}

void
_XmCvtStringToPrimBottomShadowPixmap(XrmValuePtr args, Cardinal *num_args,
				     XrmValuePtr fromVal, XrmValuePtr toVal)
{
    DEBUGOUT(XdbDebug(__FILE__, NULL, "_XmCvtStringToPrimBottomShadowPixmap\n"));

    BuildPixmap(XtNumber(PrimBottomShadowPixmapArgs),
		"_XmCvtStringToPrimBottomShadowPixmap",
		args, num_args, fromVal, toVal);
}

void
_XmCvtStringToManForegroundPixmap(XrmValuePtr args, Cardinal *num_args,
				  XrmValuePtr fromVal, XrmValuePtr toVal)
{
    DEBUGOUT(XdbDebug(__FILE__, NULL, "_XmCvtStringToManForegroundPixmap\n"));

    BuildPixmap(XtNumber(ManForegroundPixmapArgs),
		"_XmCvtStringToManForegroundPixmap",
		args, num_args, fromVal, toVal);
}

void
_XmCvtStringToManHighlightPixmap(XrmValuePtr args, Cardinal *num_args,
				 XrmValuePtr fromVal, XrmValuePtr toVal)
{
    DEBUGOUT(XdbDebug(__FILE__, NULL, "_XmCvtStringToManHighlightPixmap\n"));

    BuildPixmap(XtNumber(ManHighlightPixmapArgs),
		"_XmCvtStringToManHighlightPixmap",
		args, num_args, fromVal, toVal);
}

void
_XmCvtStringToManTopShadowPixmap(XrmValuePtr args, Cardinal *num_args,
				 XrmValuePtr fromVal, XrmValuePtr toVal)
{
    DEBUGOUT(XdbDebug(__FILE__, NULL, "_XmCvtStringToManTopShadowPixmap\n"));

    BuildPixmap(XtNumber(ManTopShadowPixmapArgs),
		"_XmCvtStringToManTopShadowPixmap",
		args, num_args, fromVal, toVal);
}

void
_XmCvtStringToManBottomShadowPixmap(XrmValuePtr args, Cardinal *num_args,
				    XrmValuePtr fromVal, XrmValuePtr toVal)
{
    DEBUGOUT(XdbDebug(__FILE__, NULL, "_XmCvtStringToManBottomShadowPixmap\n"));

    BuildPixmap(XtNumber(ManBottomShadowPixmapArgs),
		"_XmCvtStringToManBottomShadowPixmap",
		args, num_args, fromVal, toVal);
}

void
_XmCvtStringToGadgetPixmap(XrmValuePtr args, Cardinal *num_args,
			   XrmValuePtr fromVal, XrmValuePtr toVal)
{
    XrmValue pargs[4];
    Cardinal pnargs = 4;
    Widget parent;

    DEBUGOUT(XdbDebug(__FILE__, NULL, "_XmCvtStringToGadgetPixmap\n"));

    parent = *((Widget *)args[0].addr);

    pargs[0].addr = (XPointer)&CoreScreen(parent);
    pargs[1].addr = (XPointer)&CoreBackground(parent);
    pargs[2].addr = (XPointer)&MGR_Foreground(parent);
    pargs[3].addr = (XPointer)&CoreDepth(parent);

    BuildPixmap(pnargs, "_XmCvtStringToGadgetPixmap",
		pargs, &pnargs, fromVal, toVal);
}

void
_XmRegisterPixmapConverters(void)
{
    static Boolean inited = False;
  
    DEBUGOUT(XdbDebug0(__FILE__, NULL, "%s:_XmRegisterPixmapConverters(%d)\n",
	__FILE__, __LINE__));

    if (inited)
    {
	return;
    }

    inited = True;

    XtSetTypeConverter(XtRString, XmRXmBackgroundPixmap,
                       CvtStringToBackgroundPixmap, backgroundArgs,
		       XtNumber(backgroundArgs), XtCacheNone, NULL);
    XtSetTypeConverter(XtRString, XmRPixmap,
                       CvtStringToBackgroundPixmap, backgroundArgs,
		       XtNumber(backgroundArgs), XtCacheNone, NULL);

    XtAddConverter(XtRString, XmRPrimForegroundPixmap,
                   _XmCvtStringToPrimForegroundPixmap, PrimForegroundPixmapArgs,
		   XtNumber(PrimForegroundPixmapArgs));
    XtAddConverter(XtRString,  XmRHighlightPixmap,
                   _XmCvtStringToPrimHighlightPixmap, PrimHighlightPixmapArgs,
		   XtNumber(PrimHighlightPixmapArgs));
    XtAddConverter(XtRString, XmRTopShadowPixmap,
                   _XmCvtStringToPrimTopShadowPixmap, PrimTopShadowPixmapArgs,
		   XtNumber(PrimTopShadowPixmapArgs));
    XtAddConverter(XtRString,  XmRBottomShadowPixmap,
                   _XmCvtStringToPrimBottomShadowPixmap,
		   PrimBottomShadowPixmapArgs,
		   XtNumber(PrimBottomShadowPixmapArgs));

    XtAddConverter(XtRString,  XmRManForegroundPixmap,
                   _XmCvtStringToManForegroundPixmap, ManForegroundPixmapArgs,
		   XtNumber(ManForegroundPixmapArgs));
    XtAddConverter(XtRString,  XmRManHighlightPixmap,
                   _XmCvtStringToManHighlightPixmap, ManHighlightPixmapArgs,
		   XtNumber(ManHighlightPixmapArgs));
    XtAddConverter(XtRString,  XmRManTopShadowPixmap,
                   _XmCvtStringToManTopShadowPixmap, ManTopShadowPixmapArgs,
		   XtNumber(ManTopShadowPixmapArgs));
    XtAddConverter(XtRString,  XmRManBottomShadowPixmap,
                   _XmCvtStringToManBottomShadowPixmap,
		   ManBottomShadowPixmapArgs,
		   XtNumber(ManBottomShadowPixmapArgs));

    XtAddConverter(XtRString,  XmRGadgetPixmap,
                   _XmCvtStringToGadgetPixmap, gadgetPixmapArgs,
		   XtNumber(gadgetPixmapArgs));

#if 0
    XtSetTypeConverter(XtRString,  XmRAnimationMask,
		       CvtStringToAnimationMask,
                       backgroundArgs, XtNumber(backgroundArgs),
		       XtCacheNone, NULL);

    XtSetTypeConverter(XtRString, XmRAnimationPixmap,
		       CvtStringToAnimationPixmap,
                       backgroundArgs, XtNumber(backgroundArgs),
		       XtCacheNone, NULL);
#endif
}

/*
 * These functions (_XmBlackPixel and _XmWhitePixel) appear in
 * libXm:Obso2_0.o in the Motif 2 libraries.  The prototypes are in
 * Xm/XmP.h but there were no implementations.
 *
 * The final XColor parameter makes no sense to me at all but the Motif
 * headers on HP-UX 9.05 specify a raw XColor (as opposed to the more
 * sensible XColor*) too so the signatures are correct.
 *
 * I don't know the proper behavior of these functions so I'm guessing.
 *
 *	--mu@echo-on.net, 1998.03.15
 */
Pixel
_XmWhitePixel(Screen *screen, Colormap colormap, XColor whitecolor)
{
    whitecolor.red   =
    whitecolor.green =
    whitecolor.blue  = USHRT_MAX;

    if (colormap == DefaultColormapOfScreen(screen) ||
	!XAllocColor(DisplayOfScreen(screen), colormap, &whitecolor))
    {
	if (colormap != DefaultColormapOfScreen(screen))
	{
	    DEBUGOUT(XdbDebug0(__FILE__, NULL, "%s:_XmWhitePixel(%d) - Could not allocate white color\n",
		__FILE__,__LINE__));
	    color_fail_warn();
        }
	whitecolor.pixel = WhitePixelOfScreen(screen);
    }
    return whitecolor.pixel;
}

Pixel
_XmBlackPixel(Screen *screen, Colormap colormap, XColor blackcolor)
{
    blackcolor.red   =
    blackcolor.green =
    blackcolor.blue  = 0;

    if (colormap == DefaultColormapOfScreen(screen) ||
	!XAllocColor(DisplayOfScreen(screen), colormap, &blackcolor))
    {
	if (colormap != DefaultColormapOfScreen(screen))
	{
	    DEBUGOUT(XdbDebug0(__FILE__, NULL, "%s:_XmBlackPixel(%d) - Could not allocate black color\n",
		__FILE__,__LINE__));
	    color_fail_warn();
        }
	blackcolor.pixel = BlackPixelOfScreen(screen);
    }
    return blackcolor.pixel;
}

/*
 * More convenient forms of _XmWhitePixel and _XmBlackPixel.
 * These can't be macros because the stupid last parameter is a raw structure
 * rather than a pointer.
 */
Pixel
_XmWhitePixelOfObject(Widget w)
{
    XColor unused;
    return _XmWhitePixel(XtScreenOfObject(w), ColormapOfObject(w), unused);
}

Pixel
_XmBlackPixelOfObject(Widget w)
{
    XColor unused;
    return _XmBlackPixel(XtScreenOfObject(w), ColormapOfObject(w), unused);
}
